/*******************************************************************************
+
+  LEDA 5.2  
+
+
+  gw_five_color_anim.c
+
+
+  Copyright (c) 1995-2007
+  by Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/

// $Revision: 1.2 $  $Date: 2007/02/25 00:47:32 $

#include <LEDA/graphics/graphwin.h>
#include <LEDA/graph/graph_alg_c.h>

using namespace leda;


class gw_fc_observer {

  GraphWin& gw;
  node_array<point> pos0; 
  event_item it[6];
  bool step;
  bool coloring;


  void hide_node(node v, bool hide_edges = true);

  void wait(string s) 
  { string msg = "\\bf\\blue " + s;
    if (step)
      step = gw.wait(msg);
    else
     { gw.message(msg);
       leda_wait(0.5);
      }
   }


public :

  void on_start(alg_five_color&);
  void on_finish(alg_five_color&);
  void on_touch_node(alg_five_color&,node);
  void on_merge_nodes(alg_five_color&,node,node);
  void on_remove_node(alg_five_color&,node);
  void on_color_node(alg_five_color&,node,int);

  gw_fc_observer(alg_five_color& p, GraphWin& _gw) : gw(_gw)
  { step = true;
    coloring = false;
    it[0] = attach(p.start_event,       *this, &gw_fc_observer::on_start);
    it[1] = attach(p.finish_event,      *this, &gw_fc_observer::on_finish);
    it[2] = attach(p.touch_node_event,  *this, &gw_fc_observer::on_touch_node);
    it[3] = attach(p.merge_nodes_event, *this, &gw_fc_observer::on_merge_nodes);
    it[4] = attach(p.remove_node_event, *this, &gw_fc_observer::on_remove_node);
    it[5] = attach(p.color_node_event,  *this, &gw_fc_observer::on_color_node);
  }

 ~gw_fc_observer() { detach(it,6); }

};



void gw_fc_observer::on_start(alg_five_color&) 
{
  graph& g = gw.get_graph();

  pos0.init(g);

  // set initial layout

  gw.set_flush(false);

  node v;
  forall_nodes(v,g) {
    pos0[v] = gw.get_position(v);
    gw.deselect(v);
    gw.set_shape(v,circle_node);
    gw.set_width(v,25);
    gw.set_border_color(v,black);
    edge e;
    forall_adj_edges(e,v) {
      gw.deselect(e);
      gw.set_style(e,solid_edge);
      gw.set_direction(e,undirected_edge);
    }
  }
  gw.set_flush(true);
  gw.redraw();
}


void gw_fc_observer::on_finish(alg_five_color&) {
  gw.message("\\bf End of Demo");
}


void gw_fc_observer::on_touch_node(alg_five_color& a,node v) 
{
  const list<node>& L = a.get_merged_nodes(v);

  forall(v,L) 
  { gw.set_select(v,true,false);
    gw.set_shape(v,square_node);
    edge e;
    forall_inout_edges(e,v) 
    { node w = gw.get_graph().opposite(v,e);
      if (gw.get_color(w) != invisible) gw.select(e);
     }
   }
}



void gw_fc_observer::hide_node(node v, bool hide_edges) 
{ if (hide_edges) 
  { edge e;
    forall_inout_edges(e,v)
    { gw.deselect(e);
      gw.set_color(e,invisible);
      gw.set_label_color(e,invisible);
     }
   }
  gw.set_select(v,false,false);
  gw.set_color(v,invisible);
  gw.set_border_color(v,invisible);
  gw.set_label_color(v,invisible);
  gw.set_label_type(v,index_label);
}



void gw_fc_observer::on_merge_nodes(alg_five_color& a, node v, node w) 
{
  const list<node>& Lv = a.get_merged_nodes(v);
  const list<node>& Lw = a.get_merged_nodes(w);

  color cv = gw.get_color(v);
  color cw = gw.get_color(w);

  node x;
  forall(x,Lv) gw.set_color(x,green2);
  forall(x,Lw) gw.set_color(x,green2);

  string s("Merging Nodes  ");
  s += gw.get_label(v) + " and " + gw.get_label(w);

  wait(s);

  const list<node>& L = a.get_merged_nodes(w);

  gw.set_position(L,gw.get_position(v));

  gw.set_color(w,cw);
  gw.set_color(v,cv);

  forall(x,Lv) gw.set_color(x,cv);

  forall(x,Lw) 
  { gw.set_color(x,cw); 
    gw.set_label(x,gw.get_label(v));
   }
}



void gw_fc_observer::on_remove_node(alg_five_color& a, node v) 
{
  const list<node>& L = a.get_merged_nodes(v);

  gw.select(v);

  string s("Removing Node %s",gw.get_label(L.head()));

  if (L.length() > 1)
  { s += "~~[ ";
    forall(v,L) s += string("%d ",index(v));
    s += "]";
   }

  wait(s);
  forall(v,L) hide_node(v);
}




void gw_fc_observer::on_color_node(alg_five_color& a, node v, int c) 
{
  static color colors[] = { yellow,red,blue,green,pink };

  if (!coloring)
  { coloring = true;
    step = true;
   }

  const list<node>& L = a.get_merged_nodes(v);

  forall(v,L)
  { gw.set_select(v,true,false);
    gw.set_color(v,ivory);
    gw.set_label_color(v,black);
    edge e;
    forall_inout_edges(e,v) 
    { node u = opposite(v,e);
      if (gw.get_color(u) != invisible)
      { gw.set_style(e,solid_edge);
        gw.set_color(e,black);
        gw.set_select(e,true);
       }
     }
   }
  
  string s("Coloring Node %s",gw.get_label(L.head()));

  if (L.length() > 1)
  { s += "~~[ ";
    forall(v,L) s += string("%d ",index(v));
    s += "]";
   }

  wait(s);

  graph& g = gw.get_graph();
  node_array<point> pos(g);

  forall_nodes(v,g) pos[v] = gw.get_position(v);

  forall(v,L) 
  { gw.set_select(v,false,false);
    gw.set_shape(v,circle_node);
    gw.set_color(v,colors[c]);
    gw.set_border_color(v,black);
    gw.set_label_color(v,gw_auto_color);
    pos[v] = pos0[v];
    edge e;
    forall_inout_edges(e,v) 
    { node u = opposite(v,e);
      if (gw.get_color(u) != invisible) gw.set_select(e,false);
     }
  }

  if (L.length() > 1) gw.set_layout(pos);
}



void draw_graph(GraphWin& gw, color col, int dx, int dy)
{
  window& W = gw.get_window();
  graph& G = gw.get_graph();

  bool fl = gw.set_flush(false);

  gw.set_node_color(invisible);
  gw.set_node_border_color(col);
  gw.set_node_label_color(col);
  gw.set_edge_color(col);

  node v;
  forall_nodes(v,G)
  { point p = gw.get_position(v);
    gw.set_position(v,p.translate(W.pix_to_real(dx),W.pix_to_real(dy)));
   }

  gw.redraw();

  gw.set_flush(fl);
}


int main() 
{
  GraphWin  gw;

  gw.set_edge_direction(undirected_edge);

  gw.display();
  gw.edit();

  double xmin = gw.get_xmin();
  double ymin = gw.get_ymin();
  double xmax = gw.get_xmax();
  double ymax = gw.get_ymax();

  gw.save_node_attributes();
  gw.save_edge_attributes();

/*
  gw.set_bg_color(grey1);

  draw_graph(gw,grey3,0,0);

  char* pm1 = gw.get_window().get_pixrect(xmin,ymin,xmax,ymax);
  gw.set_bg_pixmap(pm1,xmin,ymax);

  draw_graph(gw,white,-1,1);
*/

  draw_graph(gw,grey1,0,0);

  char* pm = gw.get_window().get_pixrect(xmin,ymin,xmax,ymax);
  gw.set_bg_pixmap(pm,xmin,ymax);

  gw.restore_node_attributes();
  gw.restore_edge_attributes();


  gw_fc_observer observer(FIVE_COLOR,gw);
  node_array<int> col(gw.get_graph());
      
  FIVE_COLOR(gw.get_graph(),col);

  gw.wait();

  return 0;
}




