/*******************************************************************************
+
+  LEDA 5.2  
+
+
+  maxflow_s.c
+
+
+  Copyright (c) 1995-2007
+  by Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/

// $Revision: 1.2 $  $Date: 2007/02/25 01:11:31 $

#define LEDA_CHECKING_OFF

#include <LEDA/graph/graph_alg.h>
#include <LEDA/graph/s_graph.h>
#include <LEDA/graph/dimacs.h>
#include <LEDA/core/b_queue.h>
#include <LEDA/core/array.h>
#include <LEDA/system/assert.h>

using namespace leda;

#if defined(LEDA_STD_IO_HEADERS)
using std::cout;
using std::flush;
using std::endl;
#endif


// static graph for (mis)using node-info fields
//static GRAPH<node,edge> NEXT;


template <class graph_t>
class level_queue {


typedef graph_t graph;
typedef typename graph_t::node node;
typedef typename graph_t::edge edge;

node*  head;
node*  stop;
node*  max;


node_array<node,graph_t> NEXT;

public:


level_queue(const graph& G, int sz)
{ assert(NEXT.use_node_data(G,nil));
  head = new node[sz];
  max  = head;
  stop = head + sz;
  node* p = head;
  while(p < stop) *p++ = nil; 
 }

~level_queue() 
 { delete[] head; }


void insert0(node v, int i)
{ // insert without updating max pointer
   node* h = head + i;
  NEXT[v] = *h;
  *h = v;
 }

void insert(node v, int i)
{ // insert and update max pointer
  node* h = head + i;
  if (h > max) max = h;
  NEXT[v] = *h;
  *h = v;
 }

node del_max()
{ // remove and return maximum (nil if list is empty) 
  while (*max == nil && max > head) max--; 
  node v = *max;
  if (v != nil) *max = NEXT[v];
  return v;
}

node find_max()
{ // return maximum (nil if list is empty) 
  while (*max == nil && max > head) max--; 
  return *max;
}

node del() { return del_max(); }

bool empty() { return find_max() == nil; }

void clear() 
{ *head = nil;
  while (max > head) *max-- =  nil; 
 }

void clear(int i)
{ // erase everything on and above level i
  node* p = head + i;
  while (max >= p) *max-- =  nil;
  if (max < head) max = head;
}

};



template<class NT,class graph_t>
inline void compute_dist_t(const graph_t& G, typename graph_t::node t, 
                           const edge_array<NT,graph_t>& flow, 
                           const edge_array<NT,graph_t>& cap, 
                           const node_array<NT,graph_t>& excess, 
                           node_array<int,graph_t>& dist, 
                           level_queue<graph_t>& U, 
                           b_queue<typename graph_t::node>& Q, 
                           array<int>& count)
{ 

 typedef typename graph_t::node node;
 typedef typename graph_t::edge edge;

  int n = G.number_of_nodes();
  Q.append(t);

  dist[t] = 0; 

  count.init(0);
  count[0] = 1;

  while ( !Q.empty() )
  { node v = Q.pop();
    int  d = dist[v] + 1;
    edge e;

    for(e = G.first_adj_edge(v); e; e = G.adj_succ(e))
    { if ( flow[e] == 0 ) continue;
      node u = target(e); 
      int& du = dist[u];
      if ( du >= n ) 
      { du = d;
        Q.append(u); count[d]++; 
        if ( excess[u] > 0 ) U.insert(u,d); 
       }
     }

    for(e = G.first_in_edge(v); e; e = G.in_succ(e))
    { if ( cap[e] == flow[e] ) continue;
      node u = source(e); 
      int& du = dist[u];
      if ( du >= n )  
      { du = d;  
        Q.append(u); count[d]++;
       if (excess[u] > 0) U.insert(u,d); 
      }
    }
  }
}


template<class NT, class graph_t>
inline void compute_dist_s(const graph_t& G, typename graph_t::node s, 
                           const edge_array<NT,graph_t>& flow,
                           const node_array<NT,graph_t>& excess, 
                           node_array<int,graph_t>& dist, 
                           level_queue<graph_t>& U, 
                           b_queue<typename graph_t::node>& Q)
{ 
 typedef typename graph_t::node node;
 typedef typename graph_t::edge edge;

  int n = G.number_of_nodes();
  int max_level = 2*n - 1;

  Q.append(s);
  dist[s] = n;

  while ( !Q.empty() )
  { node v = Q.pop();
    int  d = dist[v] + 1;
    edge e;
    for(e = G.first_adj_edge(v); e; e = G.adj_succ(e))
    { if ( flow[e] == 0 ) continue;
      node u = target(e); 
      int& du = dist[u];
      if ( du == max_level )
      { du  = d;  
        if (excess[u] > 0) U.insert(u,d);
        Q.append(u);
       }
     }
   } 
}



template<class NT, class graph_t>
NT MAX_FLOW_T1(const graph_t& G, 
              typename graph_t::node s, 
              typename graph_t::node t, 
              const edge_array<NT,graph_t>& cap, 
              edge_array<NT,graph_t>& flow, 
              int& num_pushes, 
              int& num_edge_inspections,
              int& num_relabels,
              int& num_global_relabels,
              int& num_gaps,
              float h)
{  

 typedef typename graph_t::node node;
 typedef typename graph_t::edge edge;

   if (s == t) error_handler(1,"MAXFLOW: source == sink");


  int n = G.number_of_nodes(); int max_level = 2*n - 1;
  int m = G.number_of_edges(); 

/*
  node_array<int> dist(G,0);
  node_array<NT>  excess(G,0);
*/

// use free data slots for node_arrays if available

  node_array<NT,graph_t> excess;
  assert(excess.use_node_data(G,0));

  //flow.init(G,0);
  edge e;
  forall_edges(e,G) flow[e] = 0;

  forall_out_edges(e,s) 
  { NT c = cap[e];
    if (c == 0) continue;
    node v = target(e);
    flow[e] = c;
    excess[s] -= c;
    excess[v] += c;
  } 

  level_queue<graph_t>  U(G,2*n);
  b_queue<node> Q(n);
  int phase_number = 1;
  array<int> count(n);
  list<node> S;
  int heuristic  = (int) (h*m);
  int limit_heur = heuristic;

  node_array<int,graph_t> dist;
  assert(dist.use_node_data(G,n));

  compute_dist_t(G,t,flow,cap,excess,dist,U,Q,count);

  num_relabels         = 0;
  num_pushes           = 0;
  num_edge_inspections = 0;
  num_global_relabels  = 0;
  num_gaps = 0;

  
  for(;;) 
  {
    
    node v = U.del();

    if (v == nil)
    { 
      if ( phase_number == 2 ) break; // done

      //dist.init(G,n);
      assert(dist.use_node_data(G,n));
      compute_dist_t(G,t,flow,cap,excess,dist,U,Q,count);

      node u;
      forall_nodes(u,G) 
      { if (dist[u] == n) 
        { S.append(u);
          dist[u] = max_level;
        }
      }
      phase_number = 2;
      compute_dist_s(G,s,flow,excess,dist,U,Q);
      continue;
    }
      
    if (v == t) continue;

    if (dist[v] == n && phase_number == 1) continue;
   
    NT  ev   = excess[v]; // excess of v 
    int dv   = dist[v];   // level of v
    int dmin = MAXINT;
    edge e;

    if ( dist[v] < n )
    { 
      for (e = G.first_adj_edge(v); e; e = G.adj_succ(e))
      { num_edge_inspections++;
        NT& fe = flow[e];
        NT  rc = cap[e] - fe;
        if (rc == 0) continue;
        node w = target(e);
        int dw = dist[w];
        if ( dw < dv ) // equivalent to ( dw == dv - 1 )
        { num_pushes++;
          NT& ew = excess[w];
          if (ew == 0) U.insert0(w,dw); 
          if (ev <= rc) 
          { ew += ev; fe += ev;
            ev = 0;   // stop: excess[v] exhausted
            break;
          }
          else
          { ew += rc; fe += rc;
            ev -= rc;
          }
        }
        else { if ( dw < dmin ) dmin = dw; }
      }
 }
    if ( ev > 0 )
    { 
      for (e = G.first_in_edge(v); e; e = G.in_succ(e))
      { num_edge_inspections++;
        NT& fe = flow[e];
        if (fe == 0) continue;
        node w = source(e);
        int dw = dist[w];
        if ( dw < dv ) // equivalent to ( dw == dv - 1 )
        { num_pushes++;
          NT& ew = excess[w];
          if (ew == 0) U.insert0(w,dw); 
          if (ev <= fe) 
          { fe -= ev; ew += ev;
            ev = 0;   // stop: excess[v] exhausted
            break;
          }
          else
          { ew += fe; ev -= fe;
            fe = 0;
          }
        }
        else { if ( dw < dmin ) dmin = dw; }
      }
 }

    excess[v] = ev;

    if (ev > 0) 
    { 
      if (num_edge_inspections <= limit_heur)
        {   
          num_relabels++;
          if (phase_number == 1)
          { if ( --count[dv] == 0 || dmin >= n - 1)
            { // v cannot reach t anymore 
              
              dist[v] = n; 
              if ( dmin < n )
              { Q.append(v); 
                node w,z;
                while ( !Q.empty() )
                { edge e; 
                  w = Q.pop(); num_gaps++;
                  forall_out_edges(e,w)
                  { if ( flow[e] < cap[e] && dist[z = G.target(e)] < n)
                    { Q.append(z);  
                      count[dist[z]]--; dist[z] = n;
                    }
                  }
                  forall_in_edges(e,w)
                  { if ( flow[e] > 0 && dist[z = G.source(e)] < n)
                    { Q.append(z); 
                      count[dist[z]]--; dist[z] = n;
                    }
                  }
                }
              }

            }
            else 
            { dist[v] = ++dmin; count[dmin]++;
              U.insert(v,dmin); 
            }
          }
          else // phase_number == 2
          { dist[v] = ++dmin; 
            U.insert(v,dmin); 
          }
 }
      else
        { limit_heur += heuristic;
          num_global_relabels++;
          
          U.clear();

          if (phase_number == 1)
          { //dist.init(G,n);
            assert(dist.use_node_data(G,n));
            compute_dist_t(G,t,flow,cap,excess,dist,U,Q,count);
            if ( U.empty() ) // U is empty
            { node u;
              forall_nodes(u,G) 
              { if (dist[u] == n) 
                { S.append(u);
                  dist[u] = max_level;
                }
              }
              phase_number = 2;
              compute_dist_s(G,s,flow,excess,dist,U,Q);
            }
          }
          else
          { node u; 
            forall(u,S) dist[u] = max_level;
            compute_dist_s(G,s,flow,excess,dist,U,Q);
          }

                   
        }
 }
  }


  return excess[t];
}



int main() 
{
  int node_slots = 3; // 2 //read_int("# node slots = ");
  int edge_slots = 2; //read_int("# egde slots = ");

  GRAPH<int,int> G(node_slots,edge_slots);

  node s,t;

  int n = read_int("|V| = ");
  int m = read_int("|E| = ");

  edge_array<int>& cap = G.edge_data();

  for(int j=0; j<4; j++)
  { cout << endl;
    switch (j)
    { 
      case 0: cout << "Random: Random Networks     " << flush;
              max_flow_gen_rand(G,s,t,n,m);
              break;

      case 1: cout << "CG1: Cherkassky/Goldberg 1  " << flush;
              max_flow_gen_CG1(G,s,t,n);
              break;

      case 2: cout << "CG2: Cherkassky/Goldberg 2  " << flush;
              max_flow_gen_CG2(G,s,t,n);
              break;

      case 3: cout << "AMO: Ahuja/Magnanti/Orlin   " << flush;
              max_flow_gen_AMO(G,s,t,n);
              break;
     }

    edge_array<int> flow;
    assert(flow.use_edge_data(G,0));
  
    int n = G.number_of_nodes();
    int m = G.number_of_edges();
    
    cout << "  (" << n << " nodes  " << m << " edges)" << endl;
    cout << endl;

    int pushes,relabels,updates,gap_nodes,inspects;
    
    float T;
    int f0;

    cout << "MAX_FLOW0   " << flush;
    T = used_time();
    f0 = MAX_FLOW(G,s,t,cap,flow,pushes,inspects,relabels,updates,gap_nodes,5);
    cout << string("time: %6.2f sec  f = %d",used_time(T),f0) << endl;

    cout << "            " << flush;
    cout << string("pushes:%d  relabels:%d  updates:%d  insp:%d  gap nodes: %d",
                    pushes, relabels, updates, inspects, gap_nodes) << endl;
    cout << endl;


    cout << "MAX_FLOW    " << flush;
    T = used_time();
    f0 = MAX_FLOW_T1((graph&)G,s,t,cap,flow,pushes,inspects,relabels,updates,gap_nodes,5);
    cout << string("time: %6.2f sec  f = %d",used_time(T),f0) << endl;

    cout << "            " << flush;
    cout << string("pushes:%d  relabels:%d  updates:%d  insp:%d  gap nodes: %d",
                    pushes, relabels, updates, inspects, gap_nodes) << endl;
    cout << endl;

{

  typedef s_graph<in_edges_graph,3,2> static_graph;

  static_graph G2(n,m);

  node_array<static_graph::node> sn(G);
  edge_array<static_graph::edge> se(G);

  node v;
  forall_nodes(v,G) sn[v] = G2.new_node();

  forall_nodes(v,G) 
  { G2.adj_source(sn[v]);
    edge e;
    forall_adj_edges(e,v) 
       se[e] = G2.adj_target(sn[target(e)]);
    G2.adj_end();
  }
  //G2.construct_in_lists();
  forall_nodes(v,G) 
  { edge e = G.last_in_edge(v);
    while (e)
    {  static_graph::edge x = se[e];
       static_graph::node t = G2.target(x);
       t->add_in_edge(x);
       e = G.in_pred(e);
     }
  }


  static_graph::node s2 = sn[s];
  static_graph::node t2 = sn[t];

  edge_array<int,static_graph>  cap2;
  assert(cap2.use_edge_data(G2));

  edge_array<int,static_graph>  flow2;
  assert(flow2.use_edge_data(G2));

  edge e;
  forall_edges(e,G) cap2[se[e]] = cap[e];

  cout << "MAX_FLOW    " << flush;
  T = used_time();
  f0 = MAX_FLOW_T1(G2,s2,t2,cap2,flow2,pushes,inspects,relabels,updates,gap_nodes,5) ;
  cout << string("time: %6.2f sec  f = %d",used_time(T),f0) << endl;

  cout << "            " << flush;
  cout << string("pushes:%d  relabels:%d  updates:%d  insp:%d  gap nodes: %d",
                  pushes, relabels, updates, inspects, gap_nodes) << endl;
  cout << endl;

/*
  forall_edges(e,G) 
     assert (flow[e] == flow2[se[e]]);
*/
}

}
    
 return 0;
}
